@@ -1,6 +1,6 @@
 from functools import partial, wraps
 from io import IOBase
-from typing import Any, Optional, Union, Sequence, Callable, Iterator, List
+from typing import Any, Union, Sequence, Callable, Iterator
 import copy
 import hashlib
 import itertools
@@ -15,12 +15,11 @@ import subprocess
 import time
 import uuid
 import warnings
-import yaml  # type: ignore
+import yaml
 import sys
 
-# Using numpy type hints is complicated, so skip it for now.
-import numpy as np  # type: ignore
-import pandas as pd  # type: ignore
+import numpy as np
+import pandas as pd
 
 pj = os.path.join
 
@@ -228,6 +227,74 @@ def pickle_read(fn: str):
     return pickle.load(open(fn, "rb"))
 
 
+def logspace(
+    start, stop, num=50, offset=0, log_func: Callable = np.log10, **kwds
+):
+    """
+    Like ``numpy.logspace`` but `start` and `stop` are not exponents but the
+    actual bounds.
+
+    Control the strength of the log-scale by `offset`, where we use
+    ``np.logspace(np.log10(start + offset), np.log10(stop + offset)) -
+    offset``. `offset=0` is equal to ``np.logspace(np.log10(start),
+    np.log10(stop))``. Higher values result in more evenly spaced points.
+
+    Parameters
+    ----------
+    start, stop, num, **kwds :
+        same as in ``np.logspace``
+    offset :
+        Control strength of log-scale.
+    log_func :
+        Must match `base`. Default is ``base=10`` as in ``np.logspace`` and so
+        ``log_func=np.log10``. If you want a different `base`, also provide a
+        matching `log_func`.
+
+    Examples
+    --------
+    Effect of different `offset` values:
+
+    >>> from matplotlib import pyplot as plt
+    >>> from psweep import logspace
+    >>> import numpy as np
+
+    >>> for ii,off in enumerate([1e-16,1e-3, 1,2,3]):
+    ...     x=logspace(0, 2, 20, offset=off)
+    ...     plt.plot(x, np.ones_like(x)*ii, "o-", label=f"{off}")
+
+    >>> plt.legend()
+    """
+    base = kwds.pop("base", 10.0)
+    # fmt: off
+    assert np.allclose(log_func(base), 1.0), f"log_func and {base=} don't match"
+    # fmt: on
+    return (
+        np.logspace(
+            log_func(start + offset),
+            log_func(stop + offset),
+            num=num,
+            base=base,
+            **kwds,
+        )
+        - offset
+    )
+
+
+def intspace(*args, dtype=np.int64, **kwds):
+    """Like ``np.linspace`` but round to integers.
+
+    The length of the returned array may be lower than specified by `num` if
+    rounding to ints results in duplicates.
+
+    Parameters
+    ----------
+    *args, **kwds :
+        Same as ``np.linspace``
+    """
+    assert "dtype" not in kwds, "Got 'dtype' multiple times."
+    return np.unique(np.round(np.linspace(*args, **kwds)).astype(dtype))
+
+
 # -----------------------------------------------------------------------------
 # git
 # -----------------------------------------------------------------------------
@@ -249,8 +316,8 @@ def in_git_repo():
 
 
 def git_enter(use_git: bool, always_commit=False):
-    path = os.path.basename(fullpath(os.curdir))
     if use_git:
+        path = os.path.basename(fullpath(os.curdir))
         if not in_git_repo():
             if always_commit:
                 system(
@@ -269,8 +336,8 @@ def git_enter(use_git: bool, always_commit=False):
 
 
 def git_exit(use_git: bool, df: pd.DataFrame):
-    path = os.path.basename(fullpath(os.curdir))
     if use_git and (not git_clean()):
+        path = os.path.basename(fullpath(os.curdir))
         system(
             f"{GIT_ADD_ALL}; git commit -m 'psweep: {path}: run_id={df._run_id.values[-1]}'"
         )
@@ -281,7 +348,7 @@ def git_exit(use_git: bool, df: pd.DataFrame):
 # -----------------------------------------------------------------------------
 
 
-def df_to_json(df: pd.DataFrame, **kwds) -> Optional[str]:
+def df_to_json(df: pd.DataFrame, **kwds) -> str:
     """Like `df.to_json` but with defaults for orient, date_unit, date_format,
     double_precision.
 
@@ -342,38 +409,107 @@ def df_read(fn: str, fmt="pickle", **kwds):
 
 def df_print(
     df: pd.DataFrame,
-    index=False,
-    special_cols=False,
-    cols: Sequence[str] = None,
+    index: bool = False,
+    special_cols: bool = False,
+    cols: Sequence[str] = [],
+    skip_cols: Sequence[str] = [],
 ):
-    """Print DataFrame, by default without the index and special fields such as
-    _pset_id.
+    """Print DataFrame, by default without the index and special columns such
+    as `_pset_id`.
 
-    Same logic as in bin/psweep-db2table but w/o tabulate support. Maybe
-    later.
+    Similar logic as in `bin/psweep-db2table`, w/o tabulate support but more
+    features (`skip_cols` for instance).
+
+    Column names are always sorted, so the order of names in e.g. `cols`
+    doesn't matter.
 
     Parameters
     ----------
     df : DataFrame
-    index : include DataFrame index
-    special_cols : include all special cols (_pset_id, ...)
-    cols : explicit list of cols, overrides special_cols
+    index :
+        include DataFrame index
+    special_cols :
+        include all special columns (`_pset_id` etc.)
+    cols :
+        explicit sequence of columns, overrides `special_cols` when special columns
+        are specified
+    skip_cols :
+        skip those columns instead of selecting them (like `cols` would), use
+        either this or `cols`; overrides `special_cols` when special columns
+        are specified
+
+    Examples
+    --------
+    >>> import pandas as pd
+    >>> df=pd.DataFrame(dict(a=rand(3), b=rand(3), _c=rand(3)))
+
+    >>> df
+              a         b        _c
+    0  0.373534  0.304302  0.161799
+    1  0.698738  0.589642  0.557172
+    2  0.343316  0.186595  0.822023
+
+    >>> ps.df_print(df)
+           a        b
+    0.373534 0.304302
+    0.698738 0.589642
+    0.343316 0.186595
+
+    >>> ps.df_print(df, special_cols=True)
+           a        b       _c
+    0.373534 0.304302 0.161799
+    0.698738 0.589642 0.557172
+    0.343316 0.186595 0.822023
+
+    >>> ps.df_print(df, index=True)
+              a        b
+    0  0.373534 0.304302
+    1  0.698738 0.589642
+    2  0.343316 0.186595
+
+    >>> ps.df_print(df, cols=["a"])
+           a
+    0.373534
+    0.698738
+    0.343316
+
+    >>> ps.df_print(df, cols=["a"], special_cols=True)
+           a       _c
+    0.373534 0.161799
+    0.698738 0.557172
+    0.343316 0.822023
+
+    >>> ps.df_print(df, cols=["a", "_c"])
+           a       _c
+    0.373534 0.161799
+    0.698738 0.557172
+    0.343316 0.822023
+
+    >>> ps.df_print(df, skip_cols=["a"])
+           b
+    0.304302
+    0.589642
+    0.186595
     """
-    df2str = lambda df: df.to_string(index=index)
-    if cols is not None:
-        print(df2str(df[cols]))
+    _special_cols = set(x for x in df.columns if x.startswith("_"))
+    if len(cols) > 0:
+        if len(skip_cols) > 0:
+            raise ValueError("Use either skip_cols or cols")
+        disp_cols = set(cols) | (_special_cols if special_cols else set())
     else:
-        if special_cols:
-            print(df2str(df))
-        else:
-            _cols = set(x for x in df.columns if not x.startswith("_"))
-            print(df2str(df[_cols]))
+        disp_cols = set(df.columns) - (
+            set() if special_cols else _special_cols
+        )
+        if len(skip_cols) > 0:
+            disp_cols = disp_cols - set(skip_cols)
+    disp_cols = list(disp_cols)
+    disp_cols.sort()
+    print(df[disp_cols].to_string(index=index))
 
 
-T = Union[pd.Series, pd.DataFrame, np.ndarray, List[bool]]
-
-
-def df_filter_conds(df: pd.DataFrame, conds: Sequence[T]) -> pd.DataFrame:
+def df_filter_conds(
+    df: pd.DataFrame, conds: Sequence[Sequence[bool]]
+) -> pd.DataFrame:
     """Filter DataFrame using bool arrays/Series/DataFrames in `conds`.
 
     Logical-and all bool sequences in `conds`. Same as
@@ -515,17 +651,14 @@ def pgrid(plists):
     return itr2params(itertools.product(*plists))
 
 
-def filter_same_hash(params: Sequence[dict], **kwds):
+def filter_same_hash(params: Sequence[dict], **kwds) -> Sequence[dict]:
     """Reduce params to unique psets.
 
     Parameters
     ----------
-    params : Sequence of dicts
-    kwds : passed to pset_hash()
-
-    Returns
-    -------
-    params_filt
+    params :
+    kwds :
+        passed to :func:`pset_hash`
     """
     msk = np.unique(
         [pset_hash(dct, **kwds) for dct in params], return_index=True
@@ -539,7 +672,7 @@ def stargrid(
     vary_labels: Sequence[str] = None,
     vary_label_col: str = "_vary",
     filter_dups=True,
-):
+) -> Sequence[dict]:
     """
     Helper to create a specific param sampling pattern.
 
@@ -633,7 +766,7 @@ def worker_wrapper(
     pset: dict,
     worker: Callable,
     tmpsave: bool = False,
-    verbose: bool = False,
+    verbose: Union[bool, Sequence[str]] = None,
     run_id: str = None,
     calc_dir: str = None,
     simulate: bool = False,
@@ -644,27 +777,29 @@ def worker_wrapper(
     assert calc_dir is not None
     pset_id = str(uuid.uuid4())
     _pset = copy.deepcopy(pset)
-    _time_utc = pd.Timestamp(time.time(), unit=PANDAS_TIME_UNIT)
     hash_alg = PSET_HASH_ALG
+    time_start = pd.Timestamp(time.time(), unit=PANDAS_TIME_UNIT)
     update = {
         "_run_id": run_id,
         "_pset_id": pset_id,
         "_calc_dir": calc_dir,
-        "_time_utc": _time_utc,
+        "_time_utc": time_start,
         f"_pset_{hash_alg}": pset_hash(pset, hash_alg, raise_error=False),
         "_pset_seq": pset_seq,
         "_run_seq": run_seq,
     }
     _pset.update(update)
-    # for printing only
-    df_row = pd.DataFrame([_pset])
-    if isinstance(verbose, bool) and verbose:
-        print(df_row)
-    elif is_seq(verbose):
-        print(df_row[verbose])
+    if verbose is not None:
+        df_row_print = pd.DataFrame([_pset], index=[time_start])
+        if isinstance(verbose, bool) and verbose:
+            df_print(df_row_print, index=True)
+        elif is_seq(verbose):
+            df_print(df_row_print[verbose], index=True)
+    t0 = time.time()
     if not simulate:
         _pset.update(worker(_pset))
-    df_row = pd.DataFrame([_pset], index=[_time_utc])
+    _pset["_pset_runtime"] = time.time() - t0
+    df_row = pd.DataFrame([_pset])
     if tmpsave:
         fn = pj(calc_dir, "tmpsave", run_id, pset_id + ".pk")
         df_write(df_row, fn)
@@ -674,14 +809,14 @@ def worker_wrapper(
 def run_local(
     worker: Callable,
     params: Sequence[dict],
-    df: Optional[pd.DataFrame] = None,
-    poolsize: Optional[int] = None,
+    df: pd.DataFrame = None,
+    poolsize: int = None,
     save=True,
     tmpsave=False,
     verbose: Union[bool, Sequence[str]] = False,
     calc_dir="calc",
     simulate=False,
-    database_dir: Optional[str] = None,
+    database_dir: str = None,
     database_basename="database.pk",
     backup=False,
     git=False,
@@ -698,18 +833,18 @@ def run_local(
     df : DataFrame
         append rows to this DataFrame, if None then either create new one or
         read existing database file from disk if found
-    poolsize : {bool, None}
-        | None : use serial execution
-        | int : use multiprocessing.Pool (even for ``poolsize=1``)
+    poolsize : {None, int}
+        * None : use serial execution
+        * int : use multiprocessing.Pool (even for ``poolsize=1``)
     save : bool
         save final DataFrame to ``<calc_dir>/database.pk`` (pickle format only)
     tmpsave : bool
         save results from each `pset` in `params` (the current DataFrame row) to
         ``<calc_dir>/tmpsave/<run_id>/<pset_id>.pk`` (pickle format only)
     verbose : {bool, sequence of str}
-        | bool : print the current DataFrame row
-        | sequence : list of DataFrame column names, print the row but only
-        | those columns
+        * bool : print the current DataFrame row
+        * sequence : list of DataFrame column names, print the row but only
+          those columns
     calc_dir : str
     simulate : bool
         run everything in ``<calc_dir>.simulate``, don't call `worker`, i.e. save
@@ -723,7 +858,7 @@ def run_local(
         Make backup of ``<calc_dir>`` to ``<calc_dir>.bak_<timestamp>_run_id_<_run_id>``
     git : bool
         Use ``git`` to commit all files written and changed by the current run
-        (`_run_id`). Make sure to create a `.gitignore` manually before if
+        (``_run_id``). Make sure to create a ``.gitignore`` manually before if
         needed.
     """
 
@@ -763,7 +898,7 @@ def run_local(
         run_seq_old = df._run_seq.values.max()
 
     if backup and len(df.index) > 0:
-        stamp = df.index.max().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
+        stamp = df._time_utc.max().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
         dst = f"{calc_dir}.bak_{stamp}_run_id_{df._run_id.values[-1]}"
         assert not os.path.exists(dst), (
             "backup destination {dst} exists, seems like there has been no new "
@@ -807,7 +942,7 @@ def run_local(
             results = pool.map(_worker_wrapper_partial, params)
 
     for df_row in results:
-        df = df.append(df_row, sort=False)
+        df = pd.concat((df, df_row), sort=False, ignore_index=True)
 
     if save:
         df_write(df, database_fn)
